/***************************************************************************
 *                                                                         *
 *  z64pfs.c                                                               *
 *                                                                         *
 *  Z64 PFS Up- and download image                                         *
 *  Version 1.00.17                                                        *
 *                                                                         *
 *  Ravemax/Dextrose                                                       *
 *                                                                         *
 ***************************************************************************/

#include "z64pfs.h"

/*--------------------------------------------------------------------------
   N64-specific startup-variables
 --------------------------------------------------------------------------*/

/*
 *  Stacks
 */

u64   bootStack[StackSize];
u64   idleThreadStack[StackSize];
u64   mainThreadStack[StackSize];

/*
 *  Threads
 */

OSThread   idleThread;
OSThread   mainThread;

/*
 *   Messages+queues
 */

OSMesg	      PiMessages[NumPIMsgs];
OSMesgQueue   PiMessageQ;
OSMesgQueue   retraceMessageQ;
OSMesg	      retraceMessageBuf;

/*--------------------------------------------------------------------------
   Global constants
 --------------------------------------------------------------------------*/

/*
 *  Strings
 */

const char *TurnOffN64_Str = "-[ Turn the N64 off ]-";

/*
 *  Addresses in ROM
 */

u32 *PfsIDRom    = (u32 *)0xB0200000;
u32 *PfsDescRom  = (u32 *)0xB0200004;
u32 *PfsChunkRom = (u32 *)0xB0200018;
u32 *PfsDataRom  = (u32 *)0xB0200218;

/*
 *  Misc
 */

const char PfsID[4] = "PFS1";

/*--------------------------------------------------------------------------
   Global variables
 --------------------------------------------------------------------------*/

/*
 *  N64 N-Lib stuff
 */

OSMesgQueue  contMessageQ;
OSMesg	     contMessageBuf;
OSContStatus contStatus[MAXCONTROLLERS];
OSContPad    contData[MAXCONTROLLERS];
OSPfs        pfs;
OSPfsState   pfsState[16];

/*
 *  Misc
 */

char isDownloadMode;
u8   cBuffer[CBufferSize];

/*--------------------------------------------------------------------------
   PiRomToRam
 --------------------------------------------------------------------------*/

void PiRomToRam(void *rom,void *ram,u32 size)
{
   while (IO_READ(PI_STATUS_REG) & 3);
   IO_WRITE(PI_DRAM_ADDR_REG,(u32)ram & 0x1FFFFFFF);
   IO_WRITE(PI_CART_ADDR_REG,((u32)rom | 0xB0000000) & 0x1FFFFFFF);
   IO_WRITE(PI_WR_LEN_REG,size-1);
   while (IO_READ(PI_STATUS_REG) & 3);
}

/*--------------------------------------------------------------------------
   PiRamToRom
 --------------------------------------------------------------------------*/

void PiRamToRom(void *ram,void *rom,u32 size)
{
   while (IO_READ(PI_STATUS_REG) & 3);
   IO_WRITE(PI_DRAM_ADDR_REG,(u32)ram & 0x1FFFFFFF);
   IO_WRITE(PI_CART_ADDR_REG,((u32)rom | 0xB0000000) & 0x1FFFFFFF);
   IO_WRITE(PI_RD_LEN_REG,size-1);
   while (IO_READ(PI_STATUS_REG) & 3);
}

/*--------------------------------------------------------------------------
   ErrorExit
 --------------------------------------------------------------------------*/

#define Err_NoJoystick   0
#define Err_NoMempak     1
#define Err_IOFailed     3

void ErrorExit(u8 code)
{
   #define WinErrWd       220
   #define WinErrHt       60
   #define WinErrX        ((ScreenWd-WinErrWd) >> 1)
   #define WinErrY        ((ScreenHt-WinErrHt) >> 1)
   #define ErrMsgX(txt)   (WinErrX+((WinErrWd-strlen((txt))*9) >> 1))

   const char *ErrorStr = "Error";
   const char *ErrorMsgStr[] = {
      "No joystick attached!",
      "No mempak found!",
      "I/O failed!"
   };

   DrawWindow(WinErrX,WinErrY,WinErrWd,WinErrHt,ErrorStr,Col_Red);
   DrawStrCol(ErrMsgX(ErrorMsgStr[code]),WinErrY+24,ErrorMsgStr[code],
	      Col_White);
   DrawStr(ErrMsgX(TurnOffN64_Str),WinErrY+38,TurnOffN64_Str);
   for(;;);
}

/*--------------------------------------------------------------------------
   CheckForPfsfile
 --------------------------------------------------------------------------*/

void CheckForPfsfile(void)
{
   PiRomToRam(PfsIDRom,cBuffer,4);
   isDownloadMode = TRUE;
   if (*((u32 *)cBuffer)== *((u32 *)PfsID))
      isDownloadMode = FALSE;
}

/*--------------------------------------------------------------------------
   InitBack
 --------------------------------------------------------------------------*/

void InitBack(void)
{
   const char *Z64PfsStr   = "Z64 PFS * Mode:";
   const char *UploadStr   = "Upload";
   const char *DownloadStr = "Download";

   u16 loop,loop2;

   /*
    *  Fill background
    */

   for (loop = 22; loop < ScreenHt; loop++)
      for (loop2 = 0; loop2 < ScreenWd; loop2++)
	 cfb[loop*ScreenWd+loop2] = GPACK_RGBA5551(0,loop2 >> 1,loop,1);

   /*
    *  Draw titlebox
    */

   DrawHLine(0,21,ScreenWd,Col_White);
   DrawFilledBox(0,0,ScreenWd,21,Col_Black);
   DrawStrCol(11,12,Z64PfsStr,Col_Yellow);

   if (isDownloadMode)
      DrawStr(146,12,DownloadStr);
   else
      DrawStr(146,12,UploadStr);

   /*
    *  Draw 16 textboxes
    */

   for (loop = 33; loop <= 213; loop+=12)
   {
      DrawFilledBox(47,loop,146,9,Col_Gray3);
      DrawFilledBox(195,loop,38,9,Col_White);
      DrawFilledBox(235,loop,38,9,Col_Gray3);
   }

   osViBlack(FALSE);
}

/*--------------------------------------------------------------------------
   InitMempak
 --------------------------------------------------------------------------*/

void InitMempak(void)
{
   OSMesgQueue serialMessageQ;
   OSMesg      serialMessageBuf;
   u8 	       pattern;
   s32	       ret;

   /*
    *  Create serial-message queue and check for controller A
    */

   osCreateMesgQueue(&serialMessageQ,&serialMessageBuf,1);
   osSetEventMesg(OS_EVENT_SI,&serialMessageQ,(OSMesg)1);
   osContInit(&serialMessageQ,&pattern,contStatus);

   if (!(pattern & 1) || (contStatus[0].errno == CONT_NO_RESPONSE_ERROR))
      ErrorExit(Err_NoJoystick);

   /*
    *  Create controller-message queue
    */

   osCreateMesgQueue(&contMessageQ,&contMessageBuf,1);
   osSetEventMesg(OS_EVENT_SI,&contMessageQ,(OSMesg)0);

   /*
    *  Check if mempak attached
    */

   osPfsIsPlug(&contMessageQ,&pattern);
   if (!(pattern & 1) ||
       ((ret = osPfsInit(&contMessageQ,&pfs,0)) == PFS_ERR_ID_FATAL) ||
	(ret == PFS_ERR_DEVICE))
      ErrorExit(Err_NoMempak);
}

/*--------------------------------------------------------------------------
   PfsToAscii < von RBubba >
 --------------------------------------------------------------------------*/

char PfsToAscii(char c)
{
   const char Symbol[14] = "!\"#'*+,-./:=?@";

   if (c == 0x0F)
      c = ' ';
   else if (c < 0x10)
      c = '?';
   else if (c < 0x1A)
      c += '0'-0x10;
   else if (c < 0x34)
      c += 'A'-0x1A;
   else if (c < 0x42)
      c = Symbol[c-0x34];
   else
      c = '?';

   return(c);
}

/*--------------------------------------------------------------------------
   DrawPfsStr
 --------------------------------------------------------------------------*/

void DrawPfsStr(u16 x,u16 y,char *str,char num)
{
   char loop;

   for (loop = 0; loop < num; loop++)
   {
      if (!str[loop])
	 break;
      DrawChar(x+loop*9,y,PfsToAscii(str[loop]));
   }
   osWritebackDCacheAll();
}

/*--------------------------------------------------------------------------
   GetPfsState
 --------------------------------------------------------------------------*/

void GetPfsState(void)
{
   char loop;

   for (loop = 0; loop < 16; loop++)
      if (osPfsFileState(&pfs,loop,&pfsState[loop]))
	 pfsState[loop].file_size = 0;
}

/*--------------------------------------------------------------------------
   ListPfsState
 --------------------------------------------------------------------------*/

void ListPfsState(OSPfsState *osps)
{
   const char *UnusedStr = "Unused";

   char loop;

   SetFontColor(Col_Blue);
   for (loop = 0; loop < 16; loop++)
   {
      if (osps[loop].file_size)
      {
	 DrawPfsStr(48,34+loop*12,osps[loop].game_name,16);
	 DrawPfsStr(196,34+loop*12,osps[loop].ext_name,4);
	 DrawU32(236,34+loop*12,osps[loop].file_size / 256,4);
      }
      else
	 DrawStr(48,34+loop*12,UnusedStr);
   }
}

/*--------------------------------------------------------------------------
   MemSet
 --------------------------------------------------------------------------*/

void *MemSet(void *buf,u8 chr,u32 size)
{
   u32 loop;

   for (loop = 0; loop < size; loop++)
      ((u8 *)buf)[loop] = chr;

   return(buf);
}

/*--------------------------------------------------------------------------
   HexView
 --------------------------------------------------------------------------*/

void HexView(void *data,u32 size)
{
   u16 posX,posY;
   u32 posS = 0;

   /*
    *  Draw back
    */

   for (posY = 21; posY < 232; posY++)
      for (posX = 8; posX < 311; posX++)
	 cfb[posY*ScreenWd+posX] = Col_Black;

   SetFontColor(Col_Aqua);

   /*
    *  Draw data
    */

   for (posY = 0; posY < 26; posY++)
   {
      SetFontColor(Col_Yellow);
      DrawU32(11,22+posY*8,posS,3);

      for (posX = 0; posX < 9; posX++)
      {
	 if (posX % 2)
	    SetFontColor(Col_Yellow)
	 else
	    SetFontColor(Col_White)

	 DrawU32(46+posX*18,22+posY*8,(u32)((u8 *)data)[posS],2);
	 DrawChar(217+posX*9,22+posY*8,((u8 *)data)[posS]);

	 posS++;
	 if (posS == size)
	 {
	    posY = 0xFFFF;
	    break;
	 }
      }
   }
}

/*--------------------------------------------------------------------------
   DownloadMempak
 --------------------------------------------------------------------------*/

void DownloadMempak(void)
{
   u32 *dataRomPos = PfsDataRom;
   s32 curFileNo;
   char ret;

   /*
    *  List mempak-content
    */

   ListPfsState(pfsState);

   /*
    *  Write header+chunks
    */

   PiRamToRom(PfsID,PfsIDRom,4);
   PiRamToRom(MemSet(cBuffer,0,20),PfsDescRom,20);
   PiRamToRom(pfsState,PfsChunkRom,512);

   /*
    *  Write data
    */

   SetFontColor(Col_Black);

   for (curFileNo = 0; curFileNo < 16; curFileNo++)
      if (pfsState[curFileNo].file_size)
      {
	 if (ret = osPfsReadWriteFile(&pfs,curFileNo,PFS_READ,(int)0,
				      pfsState[curFileNo].file_size,cBuffer))
	    ErrorExit(Err_IOFailed);
	 PiRamToRom(cBuffer,dataRomPos,pfsState[curFileNo].file_size);
	 dataRomPos += pfsState[curFileNo].file_size / sizeof(u32);
      }

   PiRomToRam(PfsDataRom+pfsState[0].file_size+pfsState[1].file_size,cBuffer,
	      pfsState[2].file_size);
   HexView(cBuffer,pfsState[2].file_size);
}

/*--------------------------------------------------------------------------
   UploadMempak
 --------------------------------------------------------------------------*/

void UploadMempak(void)
{
   // list_pfs_state(), daten-pointer als parameter
   // mempak-Inhalt mittels pfsState lschen
   // daten allocieren

   PiRomToRam(PfsChunkRom,pfsState,sizeof(OSPfsState)*16);
   ListPfsState(pfsState);
   PiRomToRam(PfsDataRom,cBuffer,256);
   HexView(cBuffer,256);
}

/*--------------------------------------------------------------------------
   CloseApp
 --------------------------------------------------------------------------*/

void CloseApp(void)
{
   const char *TransferOK_Str = "Transfer successful";

   DrawWindow(50,98,220,44,TransferOK_Str,Col_Aqua);
   DrawStrCol(62,122,TurnOffN64_Str,Col_White);
   for(;;);
}

/*--------------------------------------------------------------------------
   MainProc
 --------------------------------------------------------------------------*/

void MainProc(void *arg)
{
   CheckForPfsfile();
   InitBack();
   InitMempak();
   GetPfsState();
   if (isDownloadMode)
      DownloadMempak();
   else
      UploadMempak();
//   CloseApp();
}

/*--------------------------------------------------------------------------
   Idle
 --------------------------------------------------------------------------*/

static void Idle(void *arg)
{
   /*
    *  Video initialisation
    */

   osCreateViManager(OS_PRIORITY_VIMGR);
   osViBlack(TRUE);
   if (osTvType == OS_TV_PAL)
      osViSetMode(&osViModeTable[VideomodePAL]);
   else
      osViSetMode(&osViModeTable[VideomodeNTSC]);
   osViSwapBuffer(cfb);

   /*
    *  Create PI-manager and Retrace-message-queue
    */

   osCreatePiManager((OSPri)OS_PRIORITY_PIMGR,&PiMessageQ,PiMessages,NumPIMsgs);
   osCreateMesgQueue(&retraceMessageQ,&retraceMessageBuf,1);
   osViSetEvent(&retraceMessageQ,NULL,1);

   /*
    *  Create and start MAINPROC thread
    */

   osCreateThread(&mainThread, TID_MainProc,MainProc,(void *)0,
		  (void *)(mainThreadStack+StackSize),10);
   osStartThread(&mainThread);

   /*
    *  Idling
    */

   osSetThreadPri(0,0);
   while(1);
}

/*--------------------------------------------------------------------------
   Boot
 --------------------------------------------------------------------------*/

void boot(void)
{
   osInitialize();
   osCreateThread(&idleThread,TID_Idle,Idle,(void*)0,
		  idleThreadStack+StackSize,10);
   osStartThread(&idleThread);
}

