/*---N64_EMU.CPP---*/
#include"n64_emu.h"
#include<conio.h>
#include<bios.h>
// N64 ROM emulator I/O port addresses
const unsigned data_port         = 0x160;
const unsigned work_mode_port    = 0x162;
const unsigned address_low_port  = 0x164;
const unsigned address_high_port = 0x166;
const unsigned EEPROM_read_port  = 0x164;
// Control bit constant
const unsigned N64_EEPROM  = 0x0000;
const unsigned PC_EEPROM   = 0x1000;
const unsigned SCS_DISABLE = 0x0000;
const unsigned SCS_ENABLE  = 0x0800;
const unsigned SDIR_READ   = 0x0000;
const unsigned SDIR_WRITE  = 0x0400;
const unsigned SCLK        = 0x0200;
const unsigned SDA_WRITE   = 0x0100;
// Version 2.0 control bits
const unsigned N64_ACCESS   = 0x0000;  // Access source control, bit 1
const unsigned PC_ACCESS    = 0x0002;  // 1 = N64 / 0 = 386
const unsigned EN_EXT_CART  = 0x0000;  // Enable ext. cartridge, bit 0
const unsigned DI_EXT_CART  = 0x0001;  // 0 = Enable / 1 = Disable
const unsigned EN_LOW_16MB  = 0x2000;  // Enable LOW 16M DRAM, bit 13
const unsigned DI_LOW_16MB  = 0x0000;  // 1 = Enable / 0 = Disable
const unsigned EN_HI_16MB   = 0x4000;  // Enable HI 16M DRAM, bit 14
const unsigned DI_HI_16MB   = 0x0000;  // 1 = Enable / 0 = Disable
const unsigned EN_LO_WR     = 0x0004;  // Enable LOW DRAM writable
const unsigned DI_LO_WR     = 0x0000;  // 1 = Enable / 0 = Disable
const unsigned EN_HI_WR     = 0x0008;  // Enable HI DRAM writable
const unsigned DI_HI_WR     = 0x0000;  // 1 = Enable / 0 = Disable
const unsigned MAP_HI_HROM  = 0x0000;  // Map HI DRAM to HI ROM (0x1100)
const unsigned MAP_HI_LSRAM = 0x0040;  // Map HI DRAM to LO SRAM(0x0800)
const unsigned MAP_HI_0500  = 0x0080;  // Map HI DRAM to 0x0500
const unsigned MAP_HI_0600  = 0x00c0;  // Map HI DRAM to 0x0600
// addresses
const unsigned long ROM_LOW_ADDR  = 0x10000000;  // cartridge ROM address
const unsigned long SRAM_LOW_ADDR = 0x08000000;  // cartridge SRAM address
// N64_ROM_emulator member functions
N64_ROM_emulator::N64_ROM_emulator()
{
  m_chip_ver=1;
}
void N64_ROM_emulator::set_chip_ver(unsigned ver)
{
  m_chip_ver=ver;
}
unsigned N64_ROM_emulator::get_chip_ver()
{
  return m_chip_ver;
}
void N64_ROM_emulator::set_ctrl_reg(unsigned int work_mode)
{
  outpw(work_mode_port,work_mode);
}
void N64_ROM_emulator::set_address(unsigned long addr)
{
  unsigned char AL,AM,AH;
  unsigned int low_addr_word,high_addr_word;
  addr<<=1;
  ((unsigned char*)&low_addr_word)[1]=((unsigned char*)&addr)[0];
  ((unsigned char*)&low_addr_word)[0]=((unsigned char*)&addr)[1];
  ((unsigned char*)&high_addr_word)[1]=((unsigned char*)&addr)[2];
  ((unsigned char*)&high_addr_word)[0]=((unsigned char*)&addr)[3];
  inpw(work_mode_port);
  outpw(address_high_port,high_addr_word);
  outpw(address_low_port,low_addr_word);
}
void N64_ROM_emulator::set_access_ctrl(unsigned mode)
{
  if(m_chip_ver==1)  // chip version 1
  {
    set_ctrl_reg(mode);
    return;
  }
  else if(m_chip_ver==2)  // chip version 2
  {
    switch(mode)
    {
      case N64_ROM:
        set_ctrl_reg(N64_ACCESS|
                     EN_EXT_CART|
                     DI_LOW_16MB|
                     DI_HI_16MB);
        break;
      case N64_DRAM:
        set_ctrl_reg(N64_ACCESS|
                     EN_EXT_CART|
                     EN_LOW_16MB|
                     EN_HI_16MB|
                     EN_HI_WR|
                     MAP_HI_LSRAM);
        break;
      case PC_ROM:
        set_ctrl_reg(PC_ACCESS|
                     EN_EXT_CART|
                     DI_LOW_16MB|
                     DI_HI_16MB);
        break;
      case PC_DRAM:
        set_ctrl_reg(PC_ACCESS|
                     EN_EXT_CART|
                     EN_LOW_16MB|
                     EN_HI_16MB);
        break;
      // ver 2.0 contorl mode
      case N64_16M_ROM_SRAM:
        set_ctrl_reg(N64_ACCESS|
                     EN_EXT_CART|
                     EN_LOW_16MB|
                     EN_HI_16MB|
                     EN_HI_WR|
                     MAP_HI_LSRAM);  // as N64_DRAM
        break;
      case N64_32M_ROM:
        set_ctrl_reg(N64_ACCESS|
                     EN_EXT_CART|
                     EN_LOW_16MB|
                     EN_HI_16MB|
                     MAP_HI_HROM);
	break;
      case USER_BIOS:
	set_ctrl_reg(N64_ACCESS|
		     DI_EXT_CART|
		     EN_LOW_16MB|
		     EN_HI_16MB|
		     EN_LO_WR|
		     EN_HI_WR);
	break;
    }
  }
}
void N64_ROM_emulator::clear_N64_req_flag()
{
  outp(0x169,0xc0);  // set bit 6,keep bit 7(HOLD)
  outp(0x169,0x80);
}
int N64_ROM_emulator::read_N64_req_flag()
{
  if(inp(0x16a)&0x20)
    return 1;
  return 0;
}
void N64_ROM_emulator::read_data_word(unsigned char far *buf,unsigned word_cnt)
{
  _asm{
    pushf
    push es
    push di
    cld
    mov cx,word_cnt
    mov dx,data_port
    les di,buf
    rep insw
    pop di
    pop es
    popf
  }
}
void N64_ROM_emulator::write_data_word(unsigned char far *buf,unsigned word_cnt)
{
  _asm{
    pushf
    push ds
    push si
    cld
    mov cx,word_cnt
    mov dx,data_port
    lds si,buf
    rep outsw
    pop si
    pop ds
    popf
  }
}
void N64_ROM_emulator::read_ROM
     (unsigned long addr,unsigned char far *buf,unsigned word_cnt)
{
  set_access_ctrl(PC_ROM);
  set_address(addr+(ROM_LOW_ADDR>>1));
  read_data_word(buf,word_cnt);
}
void N64_ROM_emulator::read_cart_SRAM
     (unsigned long addr,unsigned char far *buf,unsigned word_cnt)
{
  // need to active EEPROM CS to active SRAM power good
  set_ctrl_reg
  (PC_ACCESS|EN_EXT_CART|DI_LOW_16MB|DI_HI_16MB|
   PC_EEPROM|SCS_ENABLE);
  set_address(addr+(SRAM_LOW_ADDR>>1));
  read_data_word(buf,word_cnt);
}
void N64_ROM_emulator::write_cart_SRAM
     (unsigned long addr,unsigned char far *buf,unsigned word_cnt)
{
  set_ctrl_reg
  (PC_ACCESS|EN_EXT_CART|DI_LOW_16MB|DI_HI_16MB|
   PC_EEPROM|SCS_ENABLE);
  set_address(addr+(SRAM_LOW_ADDR>>1));
  write_data_word(buf,word_cnt);
}
void N64_ROM_emulator::read_DRAM_mapped_SRAM
     (unsigned long addr,unsigned char far *buf,unsigned word_cnt)
{
  set_ctrl_reg
  (PC_ACCESS|EN_EXT_CART|DI_LOW_16MB|EN_HI_16MB|MAP_HI_LSRAM|
   PC_EEPROM|SCS_ENABLE);
  set_address(addr+(SRAM_LOW_ADDR>>1));
  read_data_word(buf,word_cnt);
}
void N64_ROM_emulator::write_DRAM_mapped_SRAM
     (unsigned long addr,unsigned char far *buf,unsigned word_cnt)
{
  set_ctrl_reg
  (PC_ACCESS|EN_EXT_CART|DI_LOW_16MB|EN_HI_16MB|EN_HI_WR|MAP_HI_LSRAM|
   PC_EEPROM|SCS_ENABLE);
  set_address(addr+(SRAM_LOW_ADDR>>1));
  write_data_word(buf,word_cnt);
}
void N64_ROM_emulator::read_DRAM
     (unsigned long addr,unsigned char far *buf,unsigned word_cnt)
{
  set_access_ctrl(PC_DRAM);
  set_address(addr+(ROM_LOW_ADDR>>1));
  read_data_word(buf,word_cnt);
}
void N64_ROM_emulator::write_DRAM
     (unsigned long addr,unsigned char far *buf,unsigned word_cnt)
{
  set_access_ctrl(PC_DRAM);
  set_address(addr+(ROM_LOW_ADDR>>1));
  write_data_word(buf,word_cnt);
}
void N64_ROM_emulator::set_N64_access_EEPROM()
{
  set_ctrl_reg(N64_EEPROM|SDIR_WRITE|SDA_WRITE);
  wait_relay();
}
void N64_ROM_emulator::set_PC_access_EEPROM()
{
  set_ctrl_reg(PC_EEPROM|SDIR_WRITE|SDA_WRITE);
  wait_relay();
}
void N64_ROM_emulator::reset_EEPROM()
{
  int i;
  unsigned disable_mask;
  disable_mask=PC_EEPROM|SCS_DISABLE|SDIR_READ;
  for(i=0;i<128;i++)
  {
    set_ctrl_reg(disable_mask);
    set_ctrl_reg(disable_mask|SCLK);
  }
  unsigned enable_mask;
  enable_mask=PC_EEPROM|SCS_ENABLE|SDIR_READ;
  for(i=0;i<128;i++)
  {
    set_ctrl_reg(enable_mask);
    set_ctrl_reg(enable_mask|SCLK);
  }
  send_command_wait_clock();
}
void N64_ROM_emulator::send_command_wait_clock()
{
  unsigned clock_mask;
  clock_mask=PC_EEPROM|SCS_ENABLE|SDIR_READ;
  int i;
  for(i=0;i<1024;i++)
  {
    set_ctrl_reg(clock_mask);
    set_ctrl_reg(clock_mask|SCLK);
  }
}
void N64_ROM_emulator::send_EEPROM_physical_bit(int n)
{
  unsigned mask,data_mask;
  mask=PC_EEPROM|SCS_ENABLE|SDIR_WRITE;
  if(n)
    data_mask=SDA_WRITE;
  else
    data_mask=0x0000;
  // CLK=0/Data out
  set_ctrl_reg(mask|data_mask);
  // CLK=1/Data out
  set_ctrl_reg(mask|SCLK|data_mask);
}
void N64_ROM_emulator::send_EEPROM_logical_bit(int n)
{
  unsigned phy_bit;
  if(n)
    phy_bit=0x3f;
  else
    phy_bit=0x03;
  int mask;
  for(mask=0x80;mask!=0;mask>>=1)
  {
    if(mask&phy_bit)
      send_EEPROM_physical_bit(1);
    else
      send_EEPROM_physical_bit(0);
  }
}
void N64_ROM_emulator::send_EEPROM_logical_byte(unsigned char n)
{
  int mask;
  for(mask=0x80;mask!=0;mask>>=1)
  {
    if(mask&n)
      send_EEPROM_logical_bit(1);
    else
      send_EEPROM_logical_bit(0);
  }
}
int N64_ROM_emulator::recv_EEPROM_physical_bit()
{
  const unsigned EEPROM_read_mask=0x0100;
  unsigned mask;
  unsigned data_bit;
  mask=PC_EEPROM|SCS_ENABLE|SDIR_READ;
  // CLK=1
  set_ctrl_reg(mask|SCLK);
  // CLK=0/Data out
  set_ctrl_reg(mask);
  inpw(0);  // Need to wait before read
  // read
  data_bit=inpw(EEPROM_read_port);
  if(data_bit&EEPROM_read_mask)
    data_bit=1;
  else
    data_bit=0;
  return data_bit;
}
int N64_ROM_emulator::recv_EEPROM_logical_bit()
{
  int i,phy_bit,phy_byte;
  i=0;
  for(i=0;i<16;i++)
  {
    phy_bit=recv_EEPROM_physical_bit();
    if(phy_bit==0)  // Sync data stream
      break;
  }
  if(i>=16)
    return 1;       // Time out,quit
  // recv remained 7 physical bits
  phy_byte=0;
  unsigned mask;
  for(mask=0x40;mask!=0;mask>>=1)
  {
    phy_bit=recv_EEPROM_physical_bit();
    if(phy_bit!=0)
      phy_byte|=mask;
  }
  if(phy_byte>0x0f)
    return 1;
  return 0;
}
unsigned char N64_ROM_emulator::recv_EEPROM_logical_byte()
{
  unsigned mask;
  unsigned char data_byte;
  data_byte=0;
  for(mask=0x80;mask!=0;mask>>=1)
  {
    if(recv_EEPROM_logical_bit())
      data_byte|=mask;
  }
  return data_byte;
}
int N64_ROM_emulator::wait_EEPROM_command_ack()
{
  if(recv_EEPROM_logical_byte()!=0x00)
    return 1;
  if(recv_EEPROM_logical_byte()!=0x80)
    return 1;
  if(recv_EEPROM_logical_byte()!=0x00)
    return 1;
  if(recv_EEPROM_logical_bit()!=1)
    return 1;
  // Ok,set to write mode
  set_ctrl_reg(PC_EEPROM|SCS_ENABLE|SDIR_WRITE|SCLK|SDA_WRITE);
  return 0;
}
int N64_ROM_emulator::test_EEPROM()
{
  send_command_wait_clock();
  send_EEPROM_logical_byte(0);
  send_EEPROM_logical_bit(1);
  return wait_EEPROM_command_ack();
}
void N64_ROM_emulator::read_EEPROM_page(int page,unsigned char *buf)
{
  send_command_wait_clock();
  send_EEPROM_logical_byte(0x04);
  send_EEPROM_logical_byte(page);
  send_EEPROM_logical_bit(1);
  int i;
  for(i=0;i<8;i++)
  {
    buf[i]=recv_EEPROM_logical_byte();
  }
  recv_EEPROM_logical_bit();
  set_ctrl_reg(PC_EEPROM|SCS_ENABLE|SDIR_WRITE|SCLK|SDA_WRITE);
}
void N64_ROM_emulator::read_EEPROM_page(int page,unsigned char *buf,int page_cnt)
{
  int i;
  for(i=0;i<page_cnt;i++)
  {
    read_EEPROM_page(page+i,buf+i*8);
  }
}
void N64_ROM_emulator::write_EEPROM_page(int page,unsigned char *buf)
{
  // before write,wait EEPROM ready
  const wait_retry_max=30;
  int wait_retry;
  for(wait_retry=0;wait_retry<wait_retry_max;wait_retry++)
  {
    if(test_EEPROM()==0)
      break;
  }
  if(wait_retry>=wait_retry_max)
    return;  // not ready,quit
  // write EEPROM
  send_command_wait_clock();
  send_EEPROM_logical_byte(0x05);
  send_EEPROM_logical_byte(page);
  int i;
  for(i=0;i<8;i++)
  {
    send_EEPROM_logical_byte(buf[i]);
  }
  send_EEPROM_logical_bit(1);
  recv_EEPROM_logical_byte();
  recv_EEPROM_logical_bit();
  set_ctrl_reg(PC_EEPROM|SCS_ENABLE|SDIR_WRITE|SCLK|SDA_WRITE);
  send_command_wait_clock();  // Send some clocks to wait update
  // after write,wait EEPROM ready
  for(wait_retry=0;wait_retry<wait_retry_max;wait_retry++)
  {
    if(test_EEPROM()==0)
      break;
  }
}
void N64_ROM_emulator::write_EEPROM_page(int page,unsigned char *buf,int page_cnt)
{
  int i;
  for(i=0;i<page_cnt;i++)
  {
    write_EEPROM_page(page+i,buf+i*8);
  }
}
int N64_ROM_emulator::is_N64_power_on()
{
  return (inp(0x16a)&0x40)!=0;
}
int N64_ROM_emulator::test_cart()
{
  return (inp(0x16a)&0x80)!=0;
}
void N64_ROM_emulator::wait_relay()
{
  long t;
  t=biostime(0,0);
  while(biostime(0,0)<=t+4);
}
// N64_loop member functions
void N64_loop::init(unsigned long total_bytes,unsigned blk_bytes)
{
  m_total_bytes=total_bytes;
  m_blk_bytes=blk_bytes;
  m_byte_addr=0;
}
int N64_loop::not_end()
{
  if(m_byte_addr<m_total_bytes)
    return 1;
  return 0;
}
void N64_loop::next()
{
  m_byte_addr+=m_blk_bytes;
}
unsigned N64_loop::get_blk_size()
{
  return m_blk_bytes;
}
unsigned long N64_loop::get_byte_addr()
{
  return m_byte_addr;
}
unsigned long N64_loop::get_word_addr()
{
  return m_byte_addr/2;
}
