/*---Z64_FILE.CPP---*/
#include"z64_file.h"
#include"z64_gbl.h"
#include<string.h>
#include<dos.h>
#include<fcntl.h>
#include<io.h>
#include<dir.h>
#include<dirent.h>
#include<stdio.h>
#include<stdlib.h>

const char ROM_file_ext[]="ROM";
const char EEPROM_file_ext[]="EEP";
const char SRAM_file_ext[]="RAM";
const char BIOS_file_name[]="Z64BIOS";
const char BIOS_file_ext[]=".IMG";

// Z64_file_info
Z64_file_info::Z64_file_info()
{
  memset(m_title,0,20);
  m_rom_size=0;
  m_fn[0]='\0';
  m_country=' ';
  m_type = Z64;
  m_bios_ver=0;
}
void Z64_file_info::set_title(const char *buf)
{
  memcpy(m_title,buf,20);
}
const char *Z64_file_info::get_title() const
{
  return m_title;
}
void Z64_file_info::set_rom_size(unsigned long size)
{
  m_rom_size=size;
}
unsigned long Z64_file_info::get_rom_size() const
{
  return m_rom_size;
}
void Z64_file_info::set_file_name(const char *fn)
{
  strcpy(m_fn,fn);
}
const char *Z64_file_info::get_file_name() const
{
  return m_fn;
}
void Z64_file_info::set_country(char c)
{
  m_country=c;
}
char Z64_file_info::get_country() const
{
  return m_country;
}
void Z64_file_info::set_type(int type)
{
  m_type = type;
}
int Z64_file_info::get_type() const
{
  return m_type;
}
void Z64_file_info::set_bios_ver(unsigned ver)
{
  m_bios_ver=ver;
}
unsigned Z64_file_info::get_bios_ver() const
{
  return m_bios_ver;
}
// Z64_new_file
Z64_new_file::Z64_new_file()
{
  m_fn[0]='\0';
}
void Z64_new_file::set_file_name(const char *fn)
{
  strcpy(m_fn,fn);
}
const char *Z64_new_file::get_file_name()
{
  return m_fn;
}
// Z64_file_system
static const char *add_ext(Z64_file_info &info,const char *ext)
{
  static char fn[24];
  sprintf(fn,"%s.%s",info.get_file_name(),ext);
  return fn;
}
static const char *add_ext(Z64_new_file &nf,const char *ext)
{
  static char fn[24];
  sprintf(fn,"%s.%s",nf.get_file_name(),ext);
  return fn;
}
int Z64_filesys_read_title(const char *fn,Z64_file_info *info_buf)
{
  // Read game title
  const title_len=20;
  const title_offset=0x20;
  char tmp_title[title_len];
  char title[title_len];
  int fh;
  if(_dos_open(fn,O_RDONLY,&fh)!=0)
    return 1;
  unsigned nr;
  if(lseek(fh,title_offset,SEEK_SET)==-1L)
  {
    _dos_close(fh);
    return 1;
  }
  if(_dos_read(fh,tmp_title,title_len,&nr)!=0)
  {
    _dos_close(fh);
    return 1;
  }
  // Read ok
  if(info_buf->get_type()==Z64)
    memcpy(title,tmp_title,title_len);
  else if(info_buf->get_type()==V64)
    swab(tmp_title,title,title_len);
  info_buf->set_title(title);
  // if title is empty,use filename as title
  int i;
  for(i=0;i<title_len;i++)
  {
    unsigned char ch;
    ch=title[i];
    if(ch!=0 && ch!=0x20)
      break;
  }
  if(i>=title_len)
  {
    memset(title,0x20,title_len);
    memcpy(title,fn,strlen(fn));
    info_buf->set_title(title);
  }
  // get country char
  long id_pos;
  char ch;
  if(info_buf->get_type()==Z64)
    id_pos=0x3e;
  else
    id_pos=0x3f;
  if(lseek(fh,id_pos,SEEK_SET)==-1)
  {
    _dos_close(fh);
    return 1;
  }
  if(_dos_read(fh,&ch,1,&nr)!=0)
  {
    _dos_close(fh);
    return 1;
  }
  info_buf->set_country(ch);
  // set rom size
  info_buf->set_rom_size(filelength(fh));
  _dos_close(fh);
  return 0;
}
Z64_file_system::Z64_file_system()
:use_zip(0),m_rom_open(0),m_eeprom_open(0),m_bios_open(0)
{
  if(m_aspidrv.IsInstalled())
  {
    CommandInfo ci;
    SCSI_CDB_Inquiry cdb;
    char buf[96];
    cdb.LUN=0;
    cdb.DataAllocLen=sizeof(buf);
    ci.SetThread(0,1,0);
    ci.SetCDB(&cdb,sizeof(cdb));
    ci.SetDataBuffer(buf,sizeof(buf));
    WORD ret;
    ret=m_aspidrv.SCSI_ExecuteInputData(ci);
    if(ret==0)
    {
      if(memcmp(buf+16,"ZIP 100         ",16)==0)
      {
        printf("zip driver found\n");
        if(_getdrive()==3)
        {
          printf("driver=C:\nuse zip driver\n");
          use_zip=1;
	}
      }
    }
  }
}

int Z64_file_system::is_device_ready()
{
  if(_osmajor==6 && _osminor==0)  // OpenDOS
  {
    char buf[64];
    if(getcwd(buf,sizeof(buf)))
      return 1;
    return 0;
  }
  else  // MSDOS
  {
    int drv_ok;
    drv_ok=1;
    if(use_zip)
    {
      SCSI_CDB_TestUnitReady cdb;
      CommandInfo ci;
      memset(&cdb,0,sizeof(cdb));
      ci.SetThread(0,1,0);
      ci.SetCDB(&cdb,sizeof(cdb));
      ci.SetDataBuffer(NULL,0);
      drv_ok=(m_aspidrv.SCSI_ExecuteNoData(ci)==0);
      REGS reg;
      reg.x.ax=0x0d00;
      int86(0x21,&reg,&reg);
      char buf[64];
      getcwd(buf,sizeof(buf));
    }
    if(drv_ok)
    {
      char buf[64];
      if(getcwd(buf,sizeof(buf)))
        return 1;
      return 0;
    }
    return 0;
  }
}
long Z64_file_system::get_disk_space()
{
  diskfree_t df;
  if(_dos_getdiskfree(0,&df)!=0)
    return -1;
  return (long)df.avail_clusters*(long)df.sectors_per_cluster*
	 (long)df.bytes_per_sector;
}
// Read PFS-title
int read_pfs_title(const char *fn,Z64_file_info *info_buf)
{
   const unsigned short PFSID = 0x464D;

   int pfs_fh;
   unsigned short id_check;
   char ctemp[20];
   char loop;

  if (_dos_open(fn,O_RDONLY,&pfs_fh))
     return(1);
  _read(pfs_fh,&id_check,2);
  if (PFSID != id_check)
  {
     _dos_close(pfs_fh);
     return(1);
  }
  _dos_read(pfs_fh,ctemp,20,NULL);

  if (ctemp[0] == 0)
  {
     memset(ctemp,0x20,20);
     memcpy(ctemp,fn,strlen(fn));
  }
  else
  {
     for (loop = 1; loop < 20; loop++)
	if (ctemp[loop] == 0)
	{
	   memset(ctemp+loop,0x20,20-loop);
	   break;
	}
  }

  info_buf->set_title(ctemp);
  return(0);
}

int Z64_file_system::get_file_list(Z64_file_info *info_buf,int buf_num,int &info_num)
{
  DIR *pdir;
  dirent *ent;
  info_num=0;
  pdir=opendir("");
  if(pdir==NULL)
    return 1;
  char drive[MAXDRIVE];
  char dir[MAXDIR];
  char file[MAXFILE];
  char ext[MAXEXT];
  // Search .ROM file
  while(1)
  {
    ent=readdir(pdir);
    if(ent==NULL)
      break;
    fnsplit(ent->d_name,drive,dir,file,ext);
    unsigned bios_ver;
    // ROM file
    if(stricmp(ext,".ROM")==0)
    {
      if(info_num>=buf_num)
      {
	closedir(pdir);
        return 1;
      }
      info_buf[info_num].set_type(Z64);
      // read ROM title
      if(Z64_filesys_read_title(ent->d_name,info_buf+info_num)!=0)
      {
	closedir(pdir);
	return 1;
      }
      info_buf[info_num].set_file_name(file);
      info_num++;
    }
    // V64 file
    else if(stricmp(ext,".V64")==0)
    {
      if(info_num>=buf_num)
      {
	closedir(pdir);
	return 1;
      }
      info_buf[info_num].set_type(V64);
      // read ROM title
      if(Z64_filesys_read_title(ent->d_name,info_buf+info_num)!=0)
      {
	closedir(pdir);
	return 1;
      }
      info_buf[info_num].set_file_name(file);
      info_num++;
    }
    // PFS-file
    else if (stricmp(ext,".PFS") == 0)
    {
      if (info_num>=buf_num)
      {
	closedir(pdir);
	return 1;
      }
      info_buf[info_num].set_type(PFS);
      if (!read_pfs_title(ent->d_name,info_buf+info_num))
      {
	info_buf[info_num].set_file_name(ent->d_name);
	info_num++;
      }
    }
    // Update BIOS file
    else if
    (
     stricmp(file,BIOS_file_name)==0 && stricmp(ext,BIOS_file_ext)==0 &&
     test_bios_file(&bios_ver)==0
    )
    {
      if(info_num>=buf_num)
      {
	closedir(pdir);
	return 1;
      }
      const char bios_title[]="BIOS UPDATE";
      char bios_title_buf[20];
      memset(bios_title_buf,' ',20);
      memcpy(bios_title_buf,bios_title,sizeof(bios_title)-1);
      info_buf[info_num].set_type(BIOS_file);
      info_buf[info_num].set_title(bios_title_buf);
      info_buf[info_num].set_rom_size(0);
      info_buf[info_num].set_file_name("NUL");
      info_buf[info_num].set_bios_ver(bios_ver);
      info_num++;
    }
  }
  closedir(pdir);
  return 0;
}
int Z64_file_system::open_rom_file_read(Z64_file_info &info_buf)
{
  if(m_rom_open)
    return 1;
  char fn_buf[16];
  strcpy(fn_buf,info_buf.get_file_name());
  if (info_buf.get_type()==Z64)  // Attach fileext only if ROM
    strcat(fn_buf,".ROM");
  else if (info_buf.get_type()==V64)
    strcat(fn_buf,".V64");
  if(_dos_open(fn_buf,O_RDONLY,&m_rom_fh)!=0)
    return 1;
  m_rom_open=1;
  return 0;
}
int Z64_file_system::read_rom_file(char far *buf,unsigned len,int type)///
{
  if(m_rom_open==0)
    return 1;
  unsigned nr;
  if((_dos_read(m_rom_fh,buf,len,&nr)==0) && nr==len)
  {
    // V64 file needs byte swap
    if(type==V64)
    {
      _asm{
        push es
        push ds
        push si
        push di
        les di,buf
        lds si,buf
        mov cx,len
        shr cx,1
      }
      x:
      _asm{
        lodsw
        xchg ah,al
        stosw
        loop x
        pop di
        pop si
        pop ds
        pop es
      }
    }
    return 0;
  }
  return 1;
}
void Z64_file_system::close_rom_file()
{
  if(m_rom_open)
    _dos_close(m_rom_fh);
  m_rom_open=0;
}
int Z64_file_system::is_eeprom_file_exist(Z64_file_info &info_buf)
{
  char fn_buf[16];
  sprintf(fn_buf,"%s.EEP",info_buf.get_file_name());
  int fh;
  if(_dos_open(fn_buf,O_RDONLY,&fh)!=0)
    return 0;
  _dos_close(fh);
  return 1;
}
int Z64_file_system::open_eeprom_file_read(Z64_file_info &info_buf)
{
  if(m_eeprom_open)
    return 1;
  char fn_buf[16];
  strcpy(fn_buf,info_buf.get_file_name());
  strcat(fn_buf,".EEP");
  if(_dos_open(fn_buf,O_RDONLY,&m_eeprom_fh)!=0)
    return 1;
  m_eeprom_open=1;
  return 0;
}
int Z64_file_system::read_eeprom_file(char far *buf,unsigned len)
{
  if(m_eeprom_open==0)
    return 1;
  unsigned nr;
  if((_dos_read(m_eeprom_fh,buf,len,&nr)==0) && nr==len)
    return 0;
  return 1;
}
void Z64_file_system::close_eeprom_file()
{
  if(m_eeprom_open)
    _dos_close(m_eeprom_fh);
  m_eeprom_open=0;
}
int Z64_file_system::get_new_file(Z64_new_file &nf)
{
  int i;
  char fn_buf[16];
  // Scan free file name
  for(i=1;i<=9999;i++)
  {
    int fh,ret;
    // Check ROM/EEP file
    sprintf(fn_buf,"%04d.ROM",i);
    ret=_dos_open(fn_buf,O_RDONLY,&fh);
    if(ret==0)  // File exist,try next
    {
      _dos_close(fh);
      continue;
    }
    sprintf(fn_buf,"%04d.EEP",i);
    ret=_dos_open(fn_buf,O_RDONLY,&fh);
    if(ret==0)
    {
      _dos_close(fh);
      continue;
    }
    // File is not exist,use this file
    sprintf(fn_buf,"%04d",i);
    nf.set_file_name(fn_buf);
    return 0;
  }
  // Out of file name
  return 1;
}
int Z64_file_system::open_rom_file_create(Z64_new_file &nf)
{
  if(m_rom_open)
    return 1;
  char fn_buf[16];
  sprintf(fn_buf,"%s.ROM",nf.get_file_name());
  if(_dos_creat(fn_buf,_A_NORMAL,&m_rom_fh)!=0)
    return 1;
  m_rom_open=1;
  return 0;
}
int Z64_file_system::write_rom_file(const char far *buf,unsigned len)
{
  if(m_rom_open==0)
    return 1;
  unsigned nw;
  if((_dos_write(m_rom_fh,(void far*)buf,len,&nw)==0) && nw==len)
    return 0;
  return 1;
}
int Z64_file_system::open_eeprom_file_write(Z64_file_info &info_buf)
{
  if(m_eeprom_open)
    return 1;
  char fn_buf[16];
  sprintf(fn_buf,"%s.EEP",info_buf.get_file_name());
  if(_dos_creat(fn_buf,_A_NORMAL,&m_eeprom_fh)!=0)
    return 1;
  m_eeprom_open=1;
  return 0;
}
int Z64_file_system::open_eeprom_file_create(Z64_new_file &nf)
{
  if(m_eeprom_open)
    return 1;
  char fn_buf[16];
  sprintf(fn_buf,"%s.EEP",nf.get_file_name());
  if(_dos_creat(fn_buf,_A_NORMAL,&m_eeprom_fh)!=0)
    return 1;
  m_eeprom_open=1;
  return 0;
}
int Z64_file_system::write_eeprom_file(const char far *buf,unsigned len)
{
  if(m_eeprom_open==0)
    return 1;
  unsigned nw;
  if((_dos_write(m_eeprom_fh,(void far*)buf,len,&nw)==0) && nw==len)
    return 0;
  return 1;
}
int Z64_file_system::is_sram_file_exist(Z64_file_info &info)
{
  if(dos_file::is_exist(add_ext(info,SRAM_file_ext)))
    return 1;
  return 0;
}
int Z64_file_system::open_sram_file_read(Z64_file_info &info)
{
  return m_sram_file.open_read(add_ext(info,SRAM_file_ext));
}
int Z64_file_system::open_sram_file_write(Z64_file_info &info)
{
  return m_sram_file.open_create(add_ext(info,SRAM_file_ext));
}
int Z64_file_system::open_sram_file_create(Z64_new_file &nf)
{
  return m_sram_file.open_create(add_ext(nf,SRAM_file_ext));
}
int Z64_file_system::read_sram_file(char far *buf,unsigned len)
{
  return m_sram_file.read(buf,len);
}
int Z64_file_system::write_sram_file(const char far *buf,unsigned len)
{
  return m_sram_file.write(buf,len);
}
void Z64_file_system::close_sram_file()
{
  m_sram_file.close();
}
int Z64_file_system::delete_new_file(Z64_new_file &nf)
{
  remove(add_ext(nf,ROM_file_ext));
  remove(add_ext(nf,EEPROM_file_ext));
  remove(add_ext(nf,SRAM_file_ext));
  return 0;
}
void Z64_file_system::delete_file(Z64_file_info &info)
{
  char fn_buf[16];
  // delete rom file
  if(info.get_type()!=BIOS_file)
  {
    sprintf(fn_buf,"%s",info.get_file_name());
    if(info.get_type()==Z64)
      strcat(fn_buf,".ROM");
    else
      strcat(fn_buf,".V64");
    remove(fn_buf);
    remove(add_ext(info,EEPROM_file_ext));
    remove(add_ext(info,SRAM_file_ext));
  }
  // delete BIOS file
  else
  {
    sprintf(fn_buf,"%s%s",BIOS_file_name,BIOS_file_ext);
    remove(fn_buf);
  }
}
int Z64_file_system::open_bios_file_read()
{
  if(m_bios_open)
    return 1;
  char fn_buf[16];
  sprintf(fn_buf,"%s%s",BIOS_file_name,BIOS_file_ext);
  if(_dos_open(fn_buf,O_RDONLY,&m_bios_fh)!=0)
    return 1;
  m_bios_open=1;
  return 0;
}
int Z64_file_system::open_bios_file_create()
{
  if(m_bios_open)
    return 1;
  char fn_buf[16];
  sprintf(fn_buf,"%s%s",BIOS_file_name,BIOS_file_ext);
  if(_dos_creat(fn_buf,_A_NORMAL,&m_bios_fh)!=0)
    return 1;
  m_bios_open=1;
  return 0;
}
int Z64_file_system::seek_bios_file(long offset)
{
  return lseek(m_bios_fh,offset,SEEK_SET)!=-1L;
}
int Z64_file_system::read_bios_file(char far *buf,unsigned len)
{
  if(m_bios_open==0)
    return 1;
  unsigned nr;
  if((_dos_read(m_bios_fh,buf,len,&nr)==0) && nr==len)
    return 0;
  return 1;
}
int Z64_file_system::write_bios_file(const char far *buf,unsigned len)
{
  if(m_bios_open==0)
    return 1;
  unsigned nr;
  if((_dos_write(m_bios_fh,(void far*)buf,len,&nr)==0) && nr==len)
    return 0;
  return 1;
}
int Z64_file_system::close_bios_file()
{
  if(m_bios_open)
    _dos_close(m_bios_fh);
  m_bios_open=0;
  return 0;
}
int Z64_file_system::test_bios_file(unsigned *ver)
{
  *ver=0;
  // Check BIOS file
  static unsigned char page_buf[512];
  int ret;
  ret=open_bios_file_read();
  if(ret!=0)
    return ret;
  ret=read_bios_file(page_buf,sizeof(page_buf));
  if(ret!=0)
  {
    close_bios_file();
    return ret;
  }
  // check 55 AA header
  if(page_buf[0]!=0x55 || page_buf[1]!=0xaa)
  {
    printf("bios file header error\n");
    close_bios_file();
    return 1;
  }
  // Prepape for checksum
  unsigned page_num,page_idx;
  page_num=page_buf[2];
  printf("page_num=%d\n",page_num);
  unsigned char chksum;
  chksum=0;
  for(page_idx=0;page_idx<page_num;page_idx++)
  {
    if(page_idx!=0)  // We already read page 0,so skip it
    {
      ret=read_bios_file(page_buf,sizeof(page_buf));
      if(ret!=0)
      {
        close_bios_file();
        return ret;
      }
    }
    int byte_idx;
    for(byte_idx=0;byte_idx<sizeof(page_buf);byte_idx++)
      chksum+=page_buf[byte_idx];
  }
  // Test check sum
  if(chksum!=0)
  {
    printf("BIOS checksum error(%x)\n",chksum);
    close_bios_file();
    return 1;
  }
  // Ok
  // Check file checksum
  static char bios_data_buf[4];
  static char old_bios_data_buf[4]={0xfa,0x33,0xc0,0xe7};
  seek_bios_file(0x70000);
  if(read_bios_file(bios_data_buf,4)!=0)
  {
    printf("read bios file data error\n");
    close_bios_file();
    return 1;
  }
  int old_bios_flag=0;
  if(memcmp(bios_data_buf,old_bios_data_buf,4)!=0)
  {
    printf("bios old signature error\n");
    old_bios_flag=0;
  }
  else
  {
    printf("bios old signature ok\n");
    old_bios_flag=1;
  }
  unsigned file_chksum;
  if(get_bios_chksum(&file_chksum)!=0)
  {
    printf("get checksum error\n");
    close_bios_file();
    return 1;
  }
  printf("checksum=%04x\n",file_chksum);
  if(old_bios_flag==1)
  {
    printf("get old bios ver\n");
    if(file_chksum==0x7053)
    {
      *ver=0x105;
    }
    else if(file_chksum==0xf866)
    {
      *ver=0x106;
    }
    else
    {
      printf("bad old bios\n");
      close_bios_file();
      return 1;
    }
  }
  else
  {
    printf("get new bios ver\n");
    if(file_chksum!=0)
    {
      printf("bad chksum\n");
      close_bios_file();
      return 1;
    }
    unsigned char ver_buf[2];
    seek_bios_file(0x70002);
    if(read_bios_file(ver_buf,2)!=0)
    {
      printf("read bios file ver error\n");
      close_bios_file();
      return 1;
    }
    *ver=(ver_buf[0]<<8)+ver_buf[1];
  }
  // Ok
  close_bios_file();
  printf("ver=%04x\n",*ver);
  return 0;
}
int Z64_file_system::get_bios_chksum(unsigned *sum)
{
  const long bios_size=512L*1024;
  const unsigned file_buf_size=32768;
  seek_bios_file(0);
  *sum=0;
  long ptr;
  for(ptr=0;ptr<bios_size;ptr+=file_buf_size)
  {
    if(read_bios_file(file_buf,file_buf_size)!=0)
    {
      printf("read error\n");
      return 1;
    }
    unsigned i;
    for(i=0;i<file_buf_size;i+=2)
    {
      *sum+=*(unsigned far *)(file_buf+i);
    }
  }
  return 0;
}
