/***************************************************************************
 *                                                                         *
 *  ram.cpp                                                                *
 *                                                                         *
 *  RAM-routines (ROM,DRAM,SRAM and EEPROM)     		           *
 *                                                                         *
 ***************************************************************************/

#include "z64unzip.h"

/*--------------------------------------------------------------------------
   Constants
 --------------------------------------------------------------------------*/

/*
 *  Different ports
 */

#define PortData          0x160
#define PortWorkMode      0x162
#define PortAddressLow    0x164
#define PortAddressHigh   0x166
#define PortEepromRead    0x164
#define PortCompareLow    0x180   // compare/mask are v3.0 only
#define PortMaskLow       0x182
#define PortCompareHigh   0x184
#define PortMaskHigh      0x186

/*
 *  HW 2/3 control bits
 */

#define EnExtCart   0x0000   // Bit 0  : Enable external cartridge
#define DiExtCart   0x0001   //
#define N64Access   0x0000   // Bit 1  : Access control
#define PCAccess    0x0002   //
#define EnLoWr      0x0004   // Bit 3  : Enable LOW DRAM writable
#define DiLoWr      0x0000   //
#define EnHiWr      0x0008   // Bit 4  : Enable HI DRAM writable
#define DiHiWr      0x0000   //
#define MapHiROM    0x0000   // Bit 6  : Bit Map HI DRAM to HI ROM (0x1100)
#define MapHiSram   0x0040   // 	 Map HI DRAM to LO SRAM (0x0800)
#define EnLow16Mb   0x2000   // Bit 13 : Enable Low 16Mb DRAM, bit 13
#define DiLow16Mb   0x0000   //
#define EnHi16Mb    0x4000   // Bit 14 : Enable HI 16M DRAM, bit 14
#define DiHi16Mb    0x0000   //

/*
 *  HW2/3 Access-rights
 */

const U16 Hw2AccessRights[] = {
   // N64_ROM
   (N64Access | EnExtCart | DiLow16Mb | DiHi16Mb),
   // N64_Dram
   (N64Access | EnExtCart | EnLow16Mb | EnHi16Mb | EnHiWr | MapHiSram),
   // PC_ROM
   (PCAccess  | EnExtCart | DiLow16Mb | DiHi16Mb),
   // PC_Dram
   (PCAccess  | EnExtCart | EnLow16Mb | EnHi16Mb),
   // N64_32M_ROM
   (N64Access | EnExtCart | EnLow16Mb | EnHi16Mb | MapHiROM),
   // UserBios
   (N64Access | DiExtCart | EnLow16Mb | EnHi16Mb | EnHiWr | EnLoWr),
   // PC_SramRead
   (PCAccess  | EnExtCart | DiLow16Mb | EnHi16Mb | MapHiSram),
   // PC_SramWrite
   (PCAccess  | EnExtCart | DiLow16Mb | EnHi16Mb | EnHiWr | MapHiSram)
};

/*
 *  Misc. addresses
 */

const U32 AddrRom  = 0x10000000;
const U32 AddrSram = 0x8000000;

/*
 *  Misc.
 */

const long SizeRamBuffer = 32768;

/*--------------------------------------------------------------------------
   Macros
 --------------------------------------------------------------------------*/

#define RamSetWorkMode(mode)   \
   outpw(PortWorkMode,(mode))

/*--------------------------------------------------------------------------
   RamSetMemoryAddress - Sets a memory-address
 --------------------------------------------------------------------------*/

void RamSetMemoryAddress(U32 addr)
{
   U16 temp;

   inpw(PortWorkMode);

   ((U8 *)&temp)[1] = ((U8 *)&addr)[2];
   ((U8 *)&temp)[0] = ((U8 *)&addr)[3];
   outpw(PortAddressHigh,temp);

   ((U8 *)&temp)[1] = ((U8 *)&addr)[0];
   ((U8 *)&temp)[0] = ((U8 *)&addr)[1];
   outpw(PortAddressLow,temp);
}

/*--------------------------------------------------------------------------
   RamSetMemoryRights - Enables/disables the memory-accesses
 --------------------------------------------------------------------------*/

Bool RamSetMemoryRights(U16 right)
{
   if (Z64hwVersion == 1)
   {
      if (right >= N64_32M_ROM)
	 return(False);
      else
	 RamSetWorkMode(right);
   }
   else
      RamSetWorkMode(Hw2AccessRights[right]);
   return(True);
}

/*--------------------------------------------------------------------------
   RamRWBase - Sets access right & offset (for Read/WriteBuffer)
 --------------------------------------------------------------------------*/

Bool RamRWBase(char dstID,U32 dstOfs,Bool ReadAccess)
{
   /*
    *  Set access-right
    */

   U16 mRight;

   switch(dstID)
   {
      case Ram_Dram :
	 mRight = PC_Dram;
	 break;
      case Ram_ROM :
	 if (!ReadAccess)
	    return(False);
      case Ram_SramCart :
	 mRight = PC_ROM;
	 break;
      case Ram_SramMapped :
	 mRight = (ReadAccess) ? PC_SramRead : PC_SramWrite;
	 break;
   }
   if (!RamSetMemoryRights(mRight))
      return(False);

   /*
    *  Set memory-address
    */

   U32 mOfs = dstOfs;

   if (dstID <= Ram_Dram)
      mOfs += AddrRom;
   else
      mOfs += AddrSram;

   RamSetMemoryAddress(mOfs);

   return(True);
}

/*--------------------------------------------------------------------------
   RamWriteBuffer - Writes data to s/dram
 --------------------------------------------------------------------------*/

Bool RamWriteBuffer(char dstID,U32 dstOfs,void far *buffer,U16 size)
{
   if (!RamRWBase(dstID,dstOfs,False))
      return(False);

   asm {
      push   ds
      mov    cx,size
      shr    cx,1
      mov    dx,PortData
      lds    si,buffer
      rep    outsw
      pop    ds
   }
   return(True);
}

