/***************************************************************************
 *                                                                         *
 *  z64unzip.cpp                                                           *
 *                                                                         *
 *  Z64-ZIPfile-decompressor                                               *
 *  v1.00                                                                  *
 *                                                                         *
 *  Written by Ravemax/Dextrose                                            *
 *                                                                         *
 *  Some parts are (c)opyright by Jean-loup Gailly and Mark Adler (ZLIB)   *
 *  and Gilles Vollant (MiniZip).                                          *
 *                                                                         *
 *  ---------------------------------------------------------------------  *
 *                                                                         *
 *  Supported file-extensions:                                             *
 *    - V64  -- Byte-swapped                                               *
 *    - ROM  -.                                                            *
 *    - N64   | Z64-format                                                 *
 *    - BIN  -'                                                            *
 *                                                                         *
 ***************************************************************************/

#include "z64unzip.h"

/*--------------------------------------------------------------------------
   Constants
 --------------------------------------------------------------------------*/

/*
 *  Error-codes
 */

enum {
   err_OK = 0,
   err_FileNotFound,
   err_WrongID,
   err_NoROM,
   err_NotEnoughMem,
   err_ZLIBInitFailed,
   err_KeyPressed,
   err_ReadFile,
   err_Decompression,
   err_CRC
};

/*
 *  Strings
 */

const char *const Z64OnlyStr =
   "[ This is a Z64-Program! ]\n";

const char *const PressAKeyStr =
   "- Press a button -";

const char *const ErrorStr[] = {
   "File I/O",
   "No valid zip",
   "No ROM found",
   "Mem..",
   "ZLIB-init",
   "Key pressed",
   "Read file",
   "Inflate",
   "Checksum"
};

/*
 *  Misc
 */

#define ROMExtNum 4
const char ROMExt[ROMExtNum][3] = {
   "V64",
   "ROM","N64","BIN"
};

/*--------------------------------------------------------------------------
   Global variables
 --------------------------------------------------------------------------*/

char Z64hwVersion;  // argument

int fHandle;
ZipHeader zHeader;
Bool needsByteSwap;

/*--------------------------------------------------------------------------
   SearchROM() - Opens the ZIP and searchs for a file with a ROM-extension
 --------------------------------------------------------------------------*/

U8 SearchROM(char *zipFilename)
{
   /*
    *  Open ZIP-archive and check ID
    */

   const U16 ZIPID = 0x4B50; // "PK"
   U16 IDCheck;

   if ((fHandle = open(zipFilename,O_RDONLY)) == -1)
      return(err_FileNotFound);

   _read(fHandle,&IDCheck,2);
   if (IDCheck != ZIPID)
   {
      close(fHandle);
      return(err_WrongID);
   }

   /*
    *  Read headers and search
    */

   char register verNum;
   char tempFilename[13];
   char extLoop;

   for (;;)
   {
      _read(fHandle,&verNum,1);
      if ((verNum != 3) || (eof(fHandle))) // No ROM found
      {
	 close(fHandle);
	 return(err_NoROM);
      }

      lseek(fHandle,5,SEEK_CUR); // Skip bitfield etc.
      _read(fHandle,&zHeader,sizeof(ZipHeader));
      _read(fHandle,tempFilename,zHeader.lenFilename);

      for (extLoop = 0; extLoop < ROMExtNum; extLoop++)
	 if (!memicmp(&tempFilename[zHeader.lenFilename-3],ROMExt[extLoop],3))
	 {
	    tempFilename[zHeader.lenFilename] = 0;
	    LCDCls();
	    LCDPuts(tempFilename);

	    lseek(fHandle,zHeader.lenExtra,SEEK_CUR); // Skip extra-field
	    if (!extLoop)
	       needsByteSwap = True;
	    else
	       needsByteSwap = False;
	    return(err_OK); // OK,found
	 }

      lseek(fHandle,(zHeader.compressed+zHeader.lenExtra+2),SEEK_CUR);
   }
}

/*--------------------------------------------------------------------------
   SwapBytes() - Far-Pointer version!
 --------------------------------------------------------------------------*/

void SwapBytes(U8 far* buffer,U16 size)
{
   asm {
      mov   cx,size
      shr   cx,1
      les   di,buffer
   }
   sb_nextbyte:
   asm {
      mov   ah,es:[di]
      mov   al,es:[di+1]
      mov   es:[di],ax
      add   di,2
      dec   cx
      jnz   sb_nextbyte
   }
}

/*--------------------------------------------------------------------------
   DecompressROM() - eh yep
 --------------------------------------------------------------------------*/

U8 DecompressROM(void)
{
   #define BufferSize 16384

   z_stream zstream;   // For inflate
   char *comprBuffer,*decomprBuffer;

   /*
    *  Reserve memory and init ZLib
    */

   if ((comprBuffer = (char *)malloc(BufferSize << 1)) == NULL) // 2 buffers
      return(err_NotEnoughMem);
   decomprBuffer = comprBuffer+BufferSize;

   zstream.total_out = 0;
   zstream.zalloc = (alloc_func)0;
   zstream.zfree = (free_func)0;
   zstream.opaque = (voidpf)0;

   if (inflateInit2(&zstream,-MAX_WBITS) != Z_OK)
   {
      free(comprBuffer);
      return(err_ZLIBInitFailed);
   }

   zstream.avail_in = (uInt)0;

   /*
    *  Decompress
    */

   long  comprLeft,decomprLeft;
   int   zlibStatus;
   U16   bytesInBuffer,toRead;
   U32   totalOut;
   Bytef *bufBefore;
   U8    zipStatus = err_OK;
   U32   CRC = 0;

   comprLeft = zHeader.compressed;
   decomprLeft = zHeader.decompressed;
   LCDProgressBar(0L);

   while (decomprLeft && (zipStatus == err_OK))
   {
      zstream.next_out = (Bytef*)decomprBuffer;
      if (decomprLeft > BufferSize)
	 zstream.avail_out = (U16)BufferSize;
      else
	 zstream.avail_out = (U16)decomprLeft;

      bytesInBuffer = 0;
      while (zstream.avail_out > 0)
      {
	 // Check for key
	 if (KeypadRead() & KeyEnter)
	 {
	    zipStatus = err_KeyPressed;
	    break;
	 }
	 // Read data
	 if ((!zstream.avail_in) && (comprLeft > 0))
	 {
	    if (!comprLeft) // EOF
	    {
	       zipStatus = err_ReadFile;
	       break;
	    }

	    if (comprLeft < BufferSize)
	       toRead = comprLeft;
	    else
	       toRead = BufferSize;

	    if (_read(fHandle,comprBuffer,toRead) == -1)
	    {
	       zipStatus = err_ReadFile;
	       break;
	    }
	    comprLeft -= toRead;

	    zstream.next_in = (Bytef*)comprBuffer;
	    zstream.avail_in = (U16)toRead;
	 }
	 // Decompress
	 totalOut = zstream.total_out;
	 bufBefore = zstream.next_out;

	 zlibStatus = inflate(&zstream,(int)Z_SYNC_FLUSH);

	 totalOut = zstream.total_out-totalOut;  // decompressed bytes
	 CRC = crc32(CRC,bufBefore,(U16)totalOut);
	 decomprLeft -= (long)totalOut;
	 bytesInBuffer += (U16)totalOut;

	 if (zlibStatus == Z_STREAM_END)
	    break;
	 else if (zlibStatus != Z_OK)
	 {
	    zipStatus = err_Decompression;
	    break;
	 }

      }
      // Transfer data
      if (needsByteSwap)
	 SwapBytes(decomprBuffer,bytesInBuffer);

      RamWriteBuffer(Ram_Dram,zstream.total_out,decomprBuffer,bytesInBuffer);
      LCDProgressBar(zHeader.decompressed,zstream.total_out);
   }

   /*
    *  Shutdown ZLib and free mem
    */

   inflateEnd(&zstream);
   free(comprBuffer);

   if (zipStatus == err_OK)
      if (CRC != zHeader.crc32)
	 return(err_CRC);

   return(zipStatus);
}

/*--------------------------------------------------------------------------
   main()
 --------------------------------------------------------------------------*/

int main(int argc,char *argv[])
{
   U8 retCode;

   if (argc != 3)
   {
      puts(Z64OnlyStr);
      return(1);
   }
   Z64hwVersion = argv[1][0]-'0';

   if ((retCode = SearchROM(argv[2])) == err_OK)
      retCode = DecompressROM();

   if (retCode != err_OK)
   {
      if (retCode > err_FileNotFound)
	 close(fHandle);

      LCDCls();
      LCDPrintf("Error: %s!\n",ErrorStr[retCode-1]);
      LCDPutsXY(1,2,PressAKeyStr);
      KeypadWait(KeyAnykey);
   }
   else
      RamSetMemoryRights(N64_Dram);

   return(0);
}

