/***************************************************************************
 *                                                                         *
 *  lcd.cpp                                                                *
 *                                                                         *
 *  Routines to access the LCD-display                                     *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdarg.h>
#include "lcd.h"

/*--------------------------------------------------------------------------
   Constants
 --------------------------------------------------------------------------*/

#define LCDIOBase   0x168

/*--------------------------------------------------------------------------
   LCDIsReady - Checks if a LCD-write is possible (busy)
 --------------------------------------------------------------------------*/

U8 LCDIsReady(void)
{
  asm {
     mov    dx,LCDIOBase+3
     mov    al,0x99
     out    dx,al             // A input/B output/C input
     mov    dx,LCDIOBase+1
     mov    al,2
     out    dx,al             // RS RW=1 EN=0
     inc    al
     out    dx,al             // RS RW=1 EN=1
     dec    dx
     in     al,dx             // Get Data
     mov    bl,al
     mov    al,2
     inc    dx
     out    dx,al             // RS RW=1 EN=0
     test   bl,128
     jne    lcd_not_ready
     mov    al,1
     jmp    lr_end
  }
  lcd_not_ready:
  asm {
     xor    al,al
  }
  lr_end:
}

/*--------------------------------------------------------------------------
   LCDWrite - Sends a byte to the LCD
 --------------------------------------------------------------------------*/

void LCDWrite(U8 id,U8 data)
{
   lw_ready_loop:
      LCDIsReady();
   asm {
      or    al,al
      jz    lw_ready_loop
      mov   al,id
      mov   dx,LCDIOBase+1
      out   dx,al          // RS RW=0 EN=0
      mov   bl,al
      mov   al,0x89
      mov   dx,LCDIOBase+3
      out   dx,al          // A,B output C input
      mov   dx,LCDIOBase
      mov   al,data
      out   dx,al          // A=data
      inc   dx
      mov   al,bl
      inc   al
      out   dx,al          // RS RW=0 EN=1
      dec   al
      out   dx,al          // RS RW=0 EN=0
   }
}

/*--------------------------------------------------------------------------
   LCDInit - Sets the LCD config and clears the screen
 --------------------------------------------------------------------------*/

void LCDInit(void)
{
   #ifndef USE_VGA
      LCDWrite(LCDWCmd,6);    // Set entry-mode
      LCDWrite(LCDWCmd,12);   // Set display-configuration
      LCDWrite(LCDWCmd,56);   // Enable 2nd LCD-row
   #endif
   LCDCls();
}

/*--------------------------------------------------------------------------
   LCDCls - Clears screen and sets cursor to 1,1
 --------------------------------------------------------------------------*/

void LCDCls(void)
{
   #ifndef USE_VGA
      LCDWrite(LCDWCmd,1);
      LCDGotoXY(1,1);
   #else
      clrscr();
   #endif
}

/*--------------------------------------------------------------------------
   LCDSetFont - Sets a new font
 --------------------------------------------------------------------------*/

void LCDSetFont(U8 from,U16 num,U8 *font)
{
   #ifndef USE_VGA
      char loop;

      for (; num; num--)
      {
	 for (loop = 0; loop < 8; loop++)
	 {
	    LCDSetCGRAMAddr(from*8+loop);
	    LCDWriteChar(*font);
	    font++;
	 }
	 from++;
      }
   #endif
}

/*--------------------------------------------------------------------------
   LCD_PUTS - String-output
 --------------------------------------------------------------------------*/

void LCDPuts(const char *str)
{
   while (*str)
   {
      LCDWriteChar(*str);
      str++;
   }
}

/*--------------------------------------------------------------------------
   LCDPutsXY - String-output at x,y
 --------------------------------------------------------------------------*/

void LCDPutsXY(U8 x,U8 y,const char *str)
{
   LCDGotoXY(x,y);
   LCDPuts(str);
}

/*--------------------------------------------------------------------------
   LCDPrintf - LCD-version of "printf()"
 --------------------------------------------------------------------------*/

void LCDPrintf(const char *fmt,...)
{
   char ctemp[21];
   va_list argPtr;

   va_start(argPtr,fmt);
   vsprintf(ctemp,fmt,argPtr);
   va_end(argPtr);
   LCDPuts(ctemp);
}

/*--------------------------------------------------------------------------
   LCDProgressBar - Really simple
 --------------------------------------------------------------------------*/

void LCDProgressBar(long base,long valueNow)
{
   U8 ProgFont[] = {
      0x09,0x00,0x00,0x09,0x00,0x00,0x09,0x00,0x19,0x10,0x10,0x19,0x10,0x10,
      0x19,0x00,0x19,0x18,0x18,0x19,0x18,0x18,0x19,0x00,0x1D,0x1C,0x1C,0x1D,
      0x1C,0x1C,0x1D,0x00,0x1F,0x1E,0x1E,0x1F,0x1E,0x1E,0x1F,0x00,0x1F,0x1F,
      0x1F,0x1F,0x1F,0x1F,0x1F,0x00
   };

   const U8 NumChars  = 6;
   const U8 NumBlocks = 16;
   const U8 MaxWidth  = NumChars*NumBlocks;
   const char *const PercentStr = " 00%";

   U8 loop;

   if (!base)
   {
      /*
       *  Initialisation
       */

      LCDSetFont(0,6,ProgFont);
      LCDGotoXY(1,2);
      for (loop = 0; loop < NumBlocks; loop++)
	 LCDWriteChar(0);
      LCDPuts(PercentStr);
   }
   else
   {
      /*
       *  Draw progress
       */

      U8 temp,ypos,chrNum;

      temp = (valueNow*MaxWidth) / base;
      ypos = temp / NumChars;
      chrNum =temp % NumChars;

      LCDGotoXY(1,2);
      for (loop = 0; loop < ypos; loop++)
	 LCDWriteChar(5);
      if (chrNum)
	 LCDWriteChar(chrNum);
      LCDGotoXY(18,2);
      LCDPrintf("%02d",(valueNow*100) / base);
   }
}
