/***************************************************************************
 *                                                                         *
 *  listpfs.c                                                              *
 *                                                                         *
 *  Lists the content of a PFS-file                                        *
 *  v1.00                                                                  *
 *                                                                         *
 *  Ravemax/Dextrose                                                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>

/*--------------------------------------------------------------------------
   Types
 --------------------------------------------------------------------------*/

typedef unsigned short  U16;
typedef unsigned long   U32;

typedef struct _PfsChunk
{
   U32    fileSize;
   U32    gameCode;
   U16    companyCode;
   char   extName[4];
   char   gameName[16];
   U16    _alignment;
} PfsChunk;

/*--------------------------------------------------------------------------
   Constant strings
 --------------------------------------------------------------------------*/

const char *const TitleStr =
   "\nListPFS v1.00 (c) Ravemax/Dextrose\n";

const char *const UsageStr =
   "Usage: LISTPFS pfsfile\n";

const char *const Error_NoFileStr =
   "Error: '%s' not found!\n\n";

const char *const Error_WrongIdStr =
   "Error: Wrong ID!\n";

const char *const Error_CorruptFileStr =
   "\nError: File is corrupt!\n";

/*--------------------------------------------------------------------------
   CheckHeader
 --------------------------------------------------------------------------*/

void CheckHeader(FILE *file)
{
   const U32 ID = 0x31534650;

   U32 idCheck;
   char desc[20],loop;

   fread(&idCheck,4,1,file);
   if (ID != idCheck)
   {
      puts(Error_WrongIdStr);
      fclose(file);
      exit(3);
   }

   fread(desc,1,20,file);
   printf("Description : ");
   for (loop = 0; loop < 20; loop++)
   {
      if (!desc[loop])
	 break;
      putchar(desc[loop]);
   }
   puts("");
}

/*--------------------------------------------------------------------------
   PfsToAscii
 --------------------------------------------------------------------------*/

char PfsToAscii(char c)
{
   const char symbol[14] = "!\"#'*+,-./:=?@";

   if (c == 0x0F)
      c = ' ';
   else if (c < 0x10)
      c = '?';
   else if (c < 0x1A)
      c += '0'-0x10;
   else if (c < 0x34)
      c += 'A'-0x1A;
   else if (c < 0x42)
      c = symbol[c-0x34];
   else
      c = '?';

   return(c);
}

/*--------------------------------------------------------------------------
   PrintPfsStr
 --------------------------------------------------------------------------*/

void PrintPfsStr(char *str,char num)
{
   char loop;

   for (loop = 0; loop < num; loop++)
   {
      if (!str[loop])
	 break;
      putchar(PfsToAscii(str[loop]));
   }
}

/*--------------------------------------------------------------------------
   SwapU16
 --------------------------------------------------------------------------*/

U16 SwapU16(U16 value)
{
   U16 temp;

   *((char *)&temp)   = *((char *)&value+1);
   *((char *)&temp+1) = *((char *)&value);

   return(temp);
}

/*--------------------------------------------------------------------------
   SwapDWord
 --------------------------------------------------------------------------*/

U32 SwapU32(U32 value)
{
   U32 temp;

   *((char *)&temp)   = *((char *)&value+3);
   *((char *)&temp+1) = *((char *)&value+2);
   *((char *)&temp+2) = *((char *)&value+1);
   *((char *)&temp+3) = *((char *)&value);

   return(temp);
}

/*--------------------------------------------------------------------------
   ListDataDesc
 --------------------------------------------------------------------------*/

void ListDataDesc(FILE *file)
{
   PfsChunk pfsc;
   char loop;

   for (loop = 0; loop < 16; loop++)
   {
      if (fread(&pfsc,1,sizeof(PfsChunk),file) != sizeof(PfsChunk))
      {
	 puts(Error_CorruptFileStr);
	 fclose(file);
	 exit(4);
      }

      if (SwapU32(pfsc.fileSize))
      {
	 printf("\n[%-2d]\n",loop);
	 printf("  Game-name      : ");
	 PrintPfsStr(pfsc.gameName,16);
	 printf("\n  Extension-name : ");
	 PrintPfsStr(pfsc.extName,4);
	 printf("\n  Filesize       : %ld Byte(s)\n",SwapU32(pfsc.fileSize));
	 printf("  Game-code      : 0x%lX\n",SwapU32(pfsc.gameCode));
	 printf("  Company-code   : 0x%X\n",SwapU16(pfsc.companyCode));
      }
   }
}

/*--------------------------------------------------------------------------
   Main
 --------------------------------------------------------------------------*/

int main(int argc,char *argv[])
{
   FILE *f;

   puts(TitleStr);
   if (argc != 2)
   {
      puts(UsageStr);
      return(1);
   }
   if ((f =  fopen(argv[1],"rb")) == NULL)
   {
      printf(Error_NoFileStr,argv[1]);
      return(2);
   }

   CheckHeader(f);
   ListDataDesc(f);

   fclose(f);
   return(0);
}

